
//-----------------------------------------------------------------------------------------
// Input and Output Structures
//-----------------------------------------------------------------------------------------
struct VSSceneIn
{
	float3 Pos	: POS;
	float3 Norm : NORMAL;
	float2 Tex	: TEXCOORD0;
};

struct GSSceneIn
{
	float4 Pos	: POS;
	float3 Norm : NORMAL;
	float2 Tex	: TEXCOORD0;
};

struct PSSceneIn
{
	float4 Pos  : SV_Position;		// SV_Position is a (S)ystem (V)ariable that denotes transformed position
	float3 Norm : TEXCOORD0;		// World transformed normal
	float2 Tex  : TEXCOORD1;
};

//-----------------------------------------------------------------------------------------
// Constant Buffers (where we store variables by frequency of update)
//-----------------------------------------------------------------------------------------
cbuffer cbEveryFrame
{
	matrix g_mWorldViewProj;
	matrix g_mWorldView;
	matrix g_mWorld;
	matrix g_mProj;
};

//-----------------------------------------------------------------------------------------
// Textures and Samplers
//-----------------------------------------------------------------------------------------
Texture2D g_txDiffuse;
sampler2D g_samLinear = sampler_state
{
	texture=g_txDiffuse;
};

//-----------------------------------------------------------------------------------------
// VertexShader: VSScene
//-----------------------------------------------------------------------------------------
GSSceneIn VSScene(VSSceneIn input)
{
	GSSceneIn output;
	
	// Transform the position into view space
	output.Pos = mul( float4(input.Pos,1.0), g_mWorldView );
	
	// Transform the Normal
	output.Norm = mul( input.Norm, (float3x3)g_mWorldView );
	
	// Pass the texcoord through
	output.Tex     = input.Tex;
	
	return output;
}

float3 GetNormal( float3 A, float3 B, float3 C )
{
	float3 AB = B - A;
	float3 AC = C - A;
	return normalize( cross(AB,AC) );
}

//-----------------------------------------------------------------------------------------
// GeometryShader 
//-----------------------------------------------------------------------------------------
[maxvertexcount(6)]
void GSScene( triangleadj GSSceneIn input[6], inout LineStream<PSSceneIn> OutputStream )
{	
	//-----------------------------------------------------------------------------------------
	// NOTE:	The input of this shader is a list of vertices, 
	//          which represent a triangle and its adjacent triangles.		
	//			The original trianle is formed by the even indices (0,2,4),
	//          according to the following figure:
	//
	//             1-----2-----3
	//              \   /X\   /
	//               \ /XXX\ /
	//                0-----4
	//                 \   /
	//                  \ / 
	//                   5
	//
	//-----------------------------------------------------------------------------------------
	
	PSSceneIn output;
	
	float3 TriNormal = GetNormal(input[0].Pos.xyz, input[2].Pos.xyz, input[4].Pos.xyz);
	float3 ViewDir = -normalize(input[0].Pos + input[2].Pos + input[4].Pos); 
	
	if (dot(TriNormal, ViewDir) > 0) {
		
	// Iterate through the adjacent vertices
	for( uint i=0; i<6; i+=2 )
	{
		uint iNextVert = (i+2)%6;
		uint iAdjVert = (i+1);
		float3 AdjNormal = GetNormal(input[i].Pos.xyz, input[iAdjVert].Pos.xyz, input[iNextVert].Pos.xyz);
		
		
		if ((dot(AdjNormal, ViewDir) <= 0) || (dot(AdjNormal, TriNormal) < 0.5)) {
		
		float4 bias = float4(0.0, 0.0, 0.03, 0.0);
			
		// append the first endpoint of the edge shared by the two triangles
		output.Pos = mul( input[i].Pos - bias, g_mProj );
		output.Norm = input[i].Norm;
		output.Tex = input[i].Tex;
		OutputStream.Append( output );
		
		// append the second endpoint of the edge shared by the two triangles
		output.Pos = mul( input[ iNextVert ].Pos - bias, g_mProj );
		output.Norm = input[ iNextVert ].Norm;
		output.Tex = input[ iNextVert ].Tex;
		OutputStream.Append( output );
		
		OutputStream.RestartStrip();
		}
		
	}
	}
}

//-----------------------------------------------------------------------------------------
// PixelShader: PSScene
//-----------------------------------------------------------------------------------------
float4 PSScene(PSSceneIn input) : SV_Target
{	
	//return a constant color
	return float4(1,0,0,1);
}

//-----------------------------------------------------------------------------------------
// Technique10: Render Lines
//-----------------------------------------------------------------------------------------
technique10 RenderLines
{
    pass p0
    {
		// Set VS, GS, and PS
        SetVertexShader( CompileShader( vs_4_0, VSScene() ) );
        SetGeometryShader( CompileShader( gs_4_0, GSScene() ) );
        SetPixelShader( CompileShader( ps_4_0, PSScene() ) );
        
    }  
}


//-----------------------------------------------------------------------------------------
// VertexShader: VSScenePiece
//-----------------------------------------------------------------------------------------
PSSceneIn VSScenePiece(VSSceneIn input)
{
	PSSceneIn output;
	
	// Transform the position into clip space
	output.Pos = mul( float4(input.Pos,1.0), g_mWorldViewProj );
	
	// Transform the Normal
	output.Norm = mul( input.Norm, (float3x3)g_mWorldView );
	
	// Pass the texcoord through
	output.Tex = input.Tex;
	
	return output;
}

//-----------------------------------------------------------------------------------------
// PixelShader: PSScenePiece
//-----------------------------------------------------------------------------------------
float4 PSScenePiece(PSSceneIn input) : SV_Target
{	
	//calculate the lighting
	float lighting = saturate( dot( normalize( input.Norm ), float3(0.0,1.0,-1.0) ));
	float4 color = lighting;//tex2D( g_samLinear, input.Tex ) * lighting;
	color.a = 0.2;
	return color;
}

//-----------------------------------------------------------------------------------------
// Technique10: RenderPiece
//-----------------------------------------------------------------------------------------
technique10 RenderSolid
{
    pass p0
    {
		// Set VS, GS, and PS
        SetVertexShader( CompileShader( vs_4_0, VSScenePiece() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PSScenePiece() ) );
        
    }  
}

